# -*- coding: utf-8 -*-

'''
:Author: Juti Noppornpitak
:Version: 1.5
:Usage: Internal

The module contains the classes used for interceptable actions/methods of any
instance of :class:`imagination.entity.Entity`.

.. note::
    Copyright (c) 2012 Juti Noppornpitak

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is furnished to do
    so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
    INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
    PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
    OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

'''

from inspect import getargspec

from imagination.decorator.validator import restrict_type, SpecialType
from imagination.meta.interception   import Interception
from imagination.meta.package        import Parameter

class EventType(object):
    '''
    Type of intercepting events
    '''

    pre_action     = 'before'
    ''' Event before the execution that doesn't care about the input given to the action. '''

    pre_condition  = 'pre'
    ''' Event before the execution that only concerns about the input given to the action. '''

    post_condition = 'post'
    ''' Event after the execution that only concerns about the returned value from the action. '''

    post_action    = 'after'
    ''' Event after the execution that doesn't care about the returned value from the action. '''

class Action(object):
    '''
    Method wrapper for intercepting actions
    '''
    @restrict_type(SpecialType.function)
    def __init__(self, f):
        self.__doc__    = f.__doc__
        self._name      = f.__name__
        self._reference = f
        self.__pre_actions  = []
        self.__post_actions = []

    def __call__(self, *args, **kwargs):
        parameters = Parameter(args, kwargs)

        self.__run_pre_events(parameters)

        feedback = self.reference(*parameters.largs, **parameters.kwargs)
        feedback = self.__run_post_events(feedback)

        return feedback

    @property
    def name(self):
        ''' Name of the action '''
        return self._name

    @property
    def reference(self):
        ''' Reference of the action '''
        return self._reference

    @restrict_type(Interception)
    def register(self, interception):
        event = str(interception.event)

        if event in [EventType.pre_action, EventType.pre_condition]:
            self.__pre_actions.append(interception)
            return
        elif event in [EventType.post_action, EventType.post_condition]:
            self.__post_actions.append(interception)
            return

        raise ValueError('The event "%s" is not recognized.' % event)

    @restrict_type(Interception)
    def __retrieve_callback(self, interception):
        callback = interception.handler.interface

        # cache callback?

        if not callable(callback):
            raise NonCallableError('%s.%s is not callable' % (
                interception.handler.id,
                interception.handler.method_name
            ))

        return callback

    @restrict_type(Parameter)
    def __run_pre_events(self, parameters):
        for interception in self.__pre_actions:
            callback = self.__retrieve_callback(interception)

            if interception.event == EventType.pre_action:
                interception.handler.engage()
                continue
            elif interception.event == EventType.pre_condition:
                callback(*parameters.largs, **parameters.kwargs)
                continue

            raise ValueError('The event "%s" is not recognized.' % interception.event)

    def __run_post_events(self, feedback):
        for interception in self.__post_actions:
            callback = self.__retrieve_callback(interception)

            if interception.event == EventType.post_action:
                interception.handler.engage()
                continue
            elif interception.event == EventType.post_condition:
                feedback = callback(feedback)
                continue

            raise ValueError('The event "%s" is not recognized.' % interception.event)

        return feedback



from kotoba.kotoba import Kotoba

from imagination.decorator.validator import restrict_type
from imagination.exception import *
from imagination.loader import Loader
from imagination.locator import Locator

class Transformer(object):
    """
    Data transformer

    .. versionadded: 1.5

    :param imagination.locator.Locator locator: the entity locator
    """
    @restrict_type(Locator)
    def __init__(self, locator):
        assert locator and isinstance(locator, Locator), "Expecting an instance of imagination.locator.Locator, one of %s was given instead." % (type(locator).__name__)

        self.__locator = locator

    def cast(self, data, kind):
        '''
        Transform the given data to the given kind.

        :param data: the data to be transform
        :param str kind: the kind of data of the transformed data
        :return: the data of the given kind
        '''

        actual_data = data.data() if isinstance(data, Kotoba) else data

        if kind == 'entity':
            actual_data = self.__locator.get(actual_data)
        elif kind == 'class':
            actual_data = Loader(actual_data).package
        elif kind == 'int':
            actual_data = int(actual_data)
        elif kind == 'float':
            actual_data = float(actual_data)
        elif kind == 'bool':
            actual_data = actual_data.capitalize()

            assert actual_data == 'True' or actual_data == 'False'

            actual_data = actual_data == 'True'
        elif kind in ['list', 'tuple', 'set']:
            assert isinstance(data, Kotoba), 'Asking for a Kotoba object, getting an instance of type {}.'.format(type(data).__name__)

            actual_data = []

            for item in data.children():
                item_type = item.attribute('type')

                actual_data.append(self.cast(item, item_type))

            if kind != 'list':
                actual_data = eval(kind)(actual_data)
        elif kind == 'dict':
            assert isinstance(data, Kotoba), 'Asking for a Kotoba object, getting an instance of type {}.'.format(type(data).__name__)

            actual_data = {}

            for item in data.children():
                item_name = item.attribute('name')
                item_type = item.attribute('type')

                actual_data[item_name] = self.cast(item, item_type)
        elif kind == 'str':
            actual_data = str(actual_data)
        elif kind not in ['str', 'unicode']:
            raise ValueError('Unknown type: {} (Given data: {})'.format(kind, data))

        return actual_data

    def locator(self):
        return self.__locator

class Interception(object):
    """ Event Interception

        :param str event: the event type
        :param str actor: the ID of the actor
        :param str intercepted_action: the intercepted method
        :param str handler: the ID of the handler (interceptor)
        :param str handling_action: the handing (intercepting) method
    """
    __self_reference_keyword = 'me'

    def __init__(self, event, actor, intercepted_action, handler, handling_action):
        self.actor   = actor == self.__self_reference_keyword and handler or actor
        self.event   = event
        self.handler = handler

        self.intercepted_action = intercepted_action
        self.handling_action    = handling_action

    def __str__(self):
        return 'Interception: %s %s.%s, %s.%s' % (
            self.event,
            self.actor,
            self.intercepted_action,
            self.handler,
            self.handling_action
        )

    def __unicode__(self):
        return u'Interception: %s %s.%s, %s.%s' % (
            self.event,
            self.actor,
            self.intercepted_action,
            self.handler,
            self.handling_action
        )

class ParameterPackage(object):
    '''
    Parameter Package represents the parameter of arguments as
    a list and a dictionary to any callable objects (e.g.,
    constructor and methods).

    :param list largs:  a list of arguments
    :param dict kwargs: a dictionary of arguments
    '''
    def __init__(self, largs=None, kwargs=None):
        self.largs  = largs  or []
        self.kwargs = kwargs or {}



'''
:Author: Juti Noppornpitak
:Availability: 1.5
:Usage: Internal

The module contains the data structure of method interception to aid the analysis
and construction of the loaders and entities based on the configuration.

.. note::
    Copyright (c) 2012 Juti Noppornpitak

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is furnished to do
    so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
    INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
    PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
    OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

'''

from imagination.decorator.validator import restrict_type
from imagination.meta.aspect         import Contact

class Interception(object):
    '''
    .. note::
        Please note that both `Interception.actor` and `Interception.handler`
        are always instances of :class:`imagination.proxy.Proxy`. This is
        intentional in order to avoid direct circular dependencies.
    '''

    static_guid = 1

    @restrict_type(str, Contact, Contact)
    def __init__(self, event, actor, handler):
        self._guid    = Interception.static_guid
        self._actor   = actor
        self._event   = event
        self._handler = handler

        Interception.static_guid += 1

    @property
    def actor(self):
        return self._actor

    @property
    def event(self):
        return self._event

    @property
    def handler(self):
        return self._handler

    @staticmethod
    def self_reference_keyword():
        return 'me'


'''
:Author: Juti Noppornpitak
:Availability: 1.5

The module contains reusable input validators.

.. note::
    Copyright (c) 2012 Juti Noppornpitak

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is furnished to do
    so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
    INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
    PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
    OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

'''

import inspect
import sys

from imagination.exception import MisplacedValidatorError

_disable_decorator = 'sphinx' in sys.modules

class SpecialType(object):
    function = 'type:function'

def restrict_type(*restricted_list, **restricted_map):
    '''
    The method decorator to validate the type of inputs given to the method.

    :param `restricted_list`: the list of types to restrict the type of each
                              parameter in the same order as the parameter given
                              to the method.
    :param `restricted_map`:  the map of types to restrict the type of each
                              parameter by name.

    When the input fails the validation, an exception of type :class:`TypeError`
    is throw.

    There are a few exceptions:
     * If the given type is ``None``, there will be no restriction.
     * If the given type is ``long``, the value of ``int`` and ``float`` are also valid.
     * If the given type is ``unicode``, the valud of ``str`` is also valid.

    .. warning:: In Imagination 1.6, types ``unicode`` and ``long`` are no longer have fallback check in order to support Python 3.3.

    .. code-block:: python

        from imagination.decorator.validator import restrict_type

        # Example on a function
        @restrict_type(unicode)
        def say(context):
            print context

        class Person(object):
            # Example on a constructor
            @restrict_type(unicode, int)
            def __init__(self, name, age):
                self.name = name
                self.age  = age

                self.__friends = []

            # Example on an instance method
            @restrict_type(Person)
            def add_friend(self, person):
                self.__friends.append(person)

    '''
    def inner_decorator(reference):
        if _disable_decorator:
            return reference

        if isinstance(reference, type) or not callable(reference):
            raise MisplacedValidatorError(
                'Can only be used with callable objects, e.g., functions, class methods, instance methods and static methods.'
            )

        params   = inspect.getargspec(reference).args
        is_class = params and params[0] == 'self'

        if is_class:
            def new_reference(self, *args, **kwargs):
                __validate_type(restricted_list, restricted_map, args, kwargs)

                return reference(self, *args, **kwargs)
        else:
            def new_reference(*args, **kwargs):
                __validate_type(restricted_list, restricted_map, args, kwargs)

                return reference(*args, **kwargs)

        new_reference.__doc__ == reference.__doc__

        return new_reference

    return inner_decorator

def __validate_type(allowed_list, allowed_dictionary, argument_list, argument_dictionary):
    allowed_list = allowed_list[:len(argument_list)]
    reference    = None

    try:
        for index in range(len(allowed_list)):
            expected_type = allowed_list[index]
            reference     = argument_list[index]

            if not expected_type:
                continue

            assert __assert_type(reference, expected_type),\
                expected_type == SpecialType.function\
                and 'function'\
                or expected_type.__name__

        for key in allowed_dictionary:
            expected_type = allowed_dictionary[key]

            if key not in argument_dictionary:
                continue

            reference = argument_dictionary[key]

            if not expected_type:
                continue

            assert __assert_type(reference, expected_type),\
                expected_type.__name__

    except AssertionError as e:
        actual_type = e.message if 'message' in dir(e) else e

        raise TypeError('Argument #%s was excepting %s but %s has been given.' % (index, actual_type, type(reference).__name__))

def __assert_type(instance, expected_type):
    list_types = [list, tuple, set]

    callable_expected = expected_type == SpecialType.function
    instance_expected = isinstance(expected_type, type)

    if not callable_expected and not instance_expected:
        raise TypeError('The expected type must be a type.')

    fallback_types = []

    # If the callable is expected, validate if it is callable.

    if callable_expected:
        return callable(instance)

    # Now, validate the type of INSTANCE.
    if expected_type in list_types:
        fallback_types.extend(list_types)

    if type(instance) == expected_type or isinstance(instance, expected_type):
        return True

    for fallback_type in fallback_types:
        if type(instance) == fallback_type:
            return True

    return False


from imagination.meta.package import Parameter
from imagination.proxy        import Proxy

class Contact(object):
    def __init__(self, proxy, method_name, parameters=None):
        self._proxy       = proxy
        self._method_name = method_name
        self._parameters  = parameters

    @property
    def id(self):
        return self._proxy.id

    @property
    def method_name(self):
        return self._method_name

    @property
    def interface(self):
        return self._proxy.load().__getattribute__(self._method_name)

    @property
    def parameters(self):
        return self._parameters

    def engage(self):
        return self.interface(*self._parameters.largs, **self._parameters.kwargs)


import os
import sys

def retrieve_module(name):
    if name not in sys.modules:
        __import__(name)

    return sys.modules[name]

def retrieve_module_path(name):
    module = retrieve_module(name)

    return os.path.dirname(module.__file__) if module else None


'''
:Author: Juti Noppornpitak

The module contains the package loader used to improve code maintainability.

.. note::
    Copyright (c) 2012 Juti Noppornpitak

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is furnished to do
    so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
    INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
    PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
    OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

'''

import re
import sys
from imagination.helper import retrieve_module

class OnDemandProxy(object):
    """On-demand Proxy

    .. codeauthor:: Juti Noppornpitak <juti_n@yahoo.co.jp>
    .. versionadded:: 1.6

    .. warning:: experimental feature
    """
    def __init__(self, loader):
        self.__loader = loader

    def __call__(self, *args, **kwargs):
        return self.__loader.package

class Loader(object):
    """Package loader with lazy loading

    *path_to_package* is a string representing the package path.

    For example::

        # In this case, we load the default renderer of Tori framework.
        loader = Loader('tori.renderer.DefaultRenderer')
        # Then, instantiate the default renderer.
        renderer = loader.package('app.views')

    """
    def __init__(self, path_to_package):
        self._path         = path_to_package
        self._access_path  = re.split('\.', self._path)
        self._module_path  = '.'.join(self._access_path[:-1])
        self._module       = None
        self._package_name = self._access_path[-1]
        self._package      = None

        self.on_demand_package = OnDemandProxy(self)

    @property
    def name(self):
        ''' Get the name of the package. '''
        return self.module.__package__

    @property
    def module(self):
        ''' Get a reference to the module. '''
        if not self._module:
            self._module = retrieve_module(self._module_path)

        return self._module

    @property
    def package(self):
        ''' Get a reference to the package. '''
        if not self._package:
            self._package = self._retrieve_package()

        return self._package

    @property
    def filename(self):
        ''' Get the path to the package. '''
        return self.module.__file__

    def _retrieve_package(self):
        ''' Retrieve a package by the module path and the package name. '''

        try:
            __import__(self._module_path, fromlist=[self._package_name])
        except TypeError as exception:
            raise ImportError('Package {}.{}'.format(self._module_path, self._package_name))
        except ImportError as exception:
            raise ImportError('Could not retrieve {} from {} as {}'.format(self._package_name, self._module_path, exception.message))

        try:
            return getattr(self.module, self._package_name)
        except AttributeError as exception:
            raise ImportError('{}. (Expecting: {})'.format(exception.message, dir(self.module)))



from re import split

from kotoba.kotoba import Kotoba
from kotoba        import load_from_file

from imagination.decorator.validator import restrict_type
from imagination.entity              import Entity
from imagination.exception           import *
from imagination.loader              import Loader
from imagination.helper.data         import *
from imagination.meta.aspect         import Contact
from imagination.meta.interception   import Interception
from imagination.meta.package        import Parameter
from imagination.proxy               import Proxy

class Assembler(object):
    """
    The entity assembler via configuration files.

    .. versionadded:: 1.5

    :param imagination.helper.data.Transformer transformer: the data transformer
    """
    __known_interceptable_events = ['before', 'pre', 'post', 'after']

    @restrict_type(Transformer)
    def __init__(self, transformer):
        self.__interceptions = []
        self.__transformer   = transformer
        self.__known_proxies = {}

    def activate_passive_loading(self):
        """ Activate the passive loading mode.

            This is to delay the instantiation or forking of an entity to when
            it is referred/queried but undefined. This assumes that the queried
            entity will later be defined.

            If later on it isn't defined, the exception will be thrown.

            When this method is used, please do not forget to call :meth:`deactivate_passive_loading`
            to disable the passive mode.

            .. versionadded:: 1.7
        """
        self.locator.in_passive_mode = True

    def deactivate_passive_loading(self):
        """ Deactivate the passive loading mode.

            When :meth:`activate_passive_loading` is used, this method must be
            called to disable the passive mode.

            .. versionadded:: 1.7
        """
        self.locator.in_passive_mode = False

    def load(self, filepath):
        """
        Load the configuration.

        :param str filepath: the file path to the configuration.
        """
        xml = load_from_file(filepath)

        # First, register proxies to entities (for lazy initialization).
        for node in xml.children():
            self.__validate_node(node)
            self.__register_proxy(node)

        # Then, register loaders for entities.
        for node in xml.children():
            self.__get_interceptions(node)
            self.__register_loader(node)

        # Then, declare interceptions to target entities.
        for interception in self.__interceptions:
            self.locator\
                .get_wrapper(interception.actor.id)\
                .register_interception(interception)

        self.__interceptions = []

    @property
    def locator(self):
        """
        The injected locator via the data transformer.

        :rtype: imagination.locator.Locator
        """
        return self.__transformer.locator()

    @restrict_type(Kotoba)
    def __validate_node(self, node):
        if not node.attribute('id'):
            raise IncompatibleBlockError('Invalid entity configuration. No ID specified.')

        if not node.attribute('class'):
            raise IncompatibleBlockError('Invalid entity configuration. No class type specified.')

    @restrict_type(Kotoba)
    def __register_proxy(self, node):
        id    = node.attribute('id')
        proxy = Proxy(self.locator, id)

        self.locator.set(id, proxy)

        # this is for interceptors
        self.__known_proxies[id] = proxy

    @restrict_type(Kotoba)
    def __register_loader(self, node):
        id     = node.attribute('id')
        kind   = node.attribute('class')
        params = self.__get_params(node)
        tags   = self.__get_tags(node)

        loader = Loader(kind)

        entity = Entity(id, loader, *params.largs, **params.kwargs)

        entity.interceptable = self.__transformer.cast(node.attribute('interceptable') or 'true', 'bool')
        entity.tags          = tags

        self.locator.set(id, entity)

    @restrict_type(Kotoba)
    def __get_tags(self, node):
        tags = node.attribute('tags')

        return tags and split(' ', tags.strip()) or []

    @restrict_type(Kotoba)
    def __get_interceptions(self, node):
        for sub_node in node.children('interception'):
            self.__interceptions.append(self.__get_interception(sub_node))

    @restrict_type(Kotoba)
    def __get_interception(self, node):
        actor = None
        event = None

        intercepted_action = None
        handling_action    = None

        for given_event in self.__known_interceptable_events:
            given_actor = node.attribute(given_event)

            # If the actor is not defined, continue or if the event is already
            # set (in the earlier iteration), raise the exception.
            if not given_actor:
                continue
            elif event:
                raise MultipleInterceptingEventsWarning(given_event)

            # Initially get the name of the actor and the handler.
            actor   = given_actor
            handler = node.parent().attribute('id')

            if actor == Interception.self_reference_keyword():
                actor = handler

            # If the actor or the handler has no proxies, raise the exception.
            if actor not in self.__known_proxies:
                raise UnknownProxyError('The target (%s) of the interception is unknown.' % actor)

            if handler not in self.__known_proxies:
                raise UnknownProxyError('The handler (%s) of the interception is unknown.' % handler)

            actor   = Contact(self.__known_proxies[actor], node.attribute('do'))
            handler = Contact(self.__known_proxies[handler], node.attribute('with'), self.__get_params(node))
            event   = given_event

        return Interception(event, actor, handler)

    @restrict_type(Kotoba)
    def __get_params(self, node):
        package = Parameter()

        index = 0

        for param in node.children('param'):
            try:
                assert param.attribute('name')\
                    and param.attribute('type'),\
                    'The parameter #%d does not have either name or type.' % index
            except AssertionError as e:
                raise IncompatibleBlockError(e.message)

            index += 1
            name   = param.attribute('name')

            if name in package.kwargs:
                raise DuplicateKeyWarning('There is a parameter name "%s" with that name already registered.' % name)
                pass

            package.kwargs[name] = self.__transformer.cast(
                param,
                param.attribute('type')
            )

        return package
